@@ -1,30 +1,23 @@
 # -*- coding: utf-8 -*-
-"""
-Wandb is a library to help track machine learning experiments.
-
-For more information on wandb see https://docs.wandb.com.
+"""Use wandb to track machine learning work.
 
 The most commonly used functions/objects are:
-- wandb.init — initialize a new run at the top of your training script
-- wandb.config — track hyperparameters
-- wandb.log — log metrics over time within your training loop
-- wandb.save — save files in association with your run, like model weights
-- wandb.restore — restore the state of your code when you ran a given run
+  - wandb.init — initialize a new run at the top of your training script
+  - wandb.config — track hyperparameters and metadata
+  - wandb.log — log metrics and media over time within your training loop
 
-For examples usage, see github.com/wandb/examples
+For guides and examples, see https://docs.wandb.com/guides.
+
+For scripts and interactive notebooks, see https://github.com/wandb/examples.
+
+For reference documentation, see https://docs.wandb.com/ref/python.
 """
-
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-from __future__ import unicode_literals
-
-__version__ = "0.12.4"
+__version__ = "0.12.11"
 
 # Used with pypi checks and other messages related to pip
 _wandb_module = "wandb"
 
-import sys
+from typing import Optional
 
 from wandb.errors import Error
 
@@ -39,6 +32,8 @@ wandb.wandb_lib = wandb_sdk.lib
 
 init = wandb_sdk.init
 setup = wandb_sdk.setup
+_attach = wandb_sdk._attach
+_teardown = wandb_sdk.teardown
 save = wandb_sdk.save
 watch = wandb_sdk.watch
 unwatch = wandb_sdk.unwatch
@@ -57,8 +52,12 @@ Config = wandb_sdk.Config
 from wandb.apis import InternalApi, PublicApi
 from wandb.errors import CommError, UsageError
 
-_preinit = wandb_lib.preinit
-_lazyloader = wandb_lib.lazyloader
+_preinit = wandb.wandb_lib.preinit
+_lazyloader = wandb.wandb_lib.lazyloader
+
+# Call import module hook to setup any needed require hooks
+wandb.sdk.wandb_require._import_module_hook()
+
 from wandb import wandb_torch
 
 # Move this (keras.__init__ expects it at top level)
@@ -123,13 +122,9 @@ def _assert_is_user_process():
 # globals
 Api = PublicApi
 api = InternalApi()
-run = None
-config = _preinit.PreInitCallable(
-    _preinit.PreInitObject("wandb.config"), wandb_sdk.wandb_config.Config
-)
-summary = _preinit.PreInitCallable(
-    _preinit.PreInitObject("wandb.summary"), wandb_sdk.wandb_summary.Summary
-)
+run: Optional["wandb.sdk.wandb_run.Run"] = None
+config = _preinit.PreInitObject("wandb.config", wandb_sdk.wandb_config.Config)
+summary = _preinit.PreInitObject("wandb.summary", wandb_sdk.wandb_summary.Summary)
 log = _preinit.PreInitCallable("wandb.log", wandb_sdk.wandb_run.Run.log)
 save = _preinit.PreInitCallable("wandb.save", wandb_sdk.wandb_run.Run.save)
 restore = wandb_sdk.wandb_run.restore
@@ -163,6 +158,9 @@ tensorflow = _lazyloader.LazyLoader(
 xgboost = _lazyloader.LazyLoader(
     "wandb.xgboost", globals(), "wandb.integration.xgboost"
 )
+catboost = _lazyloader.LazyLoader(
+    "wandb.catboost", globals(), "wandb.integration.catboost"
+)
 tensorboard = _lazyloader.LazyLoader(
     "wandb.tensorboard", globals(), "wandb.integration.tensorboard"
 )
@@ -219,5 +217,4 @@ __all__ = [
     "Object3D",
     "Molecule",
     "Histogram",
-    "_enable",
 ]
