@@ -1,14 +1,11 @@
-#
-# -*- coding: utf-8 -*-
 """Handle Manager."""
 
-from __future__ import print_function
-
 import json
 import logging
 import math
 import numbers
 import os
+from queue import Queue
 from threading import Event
 import time
 from typing import (
@@ -24,16 +21,28 @@ from typing import (
     TYPE_CHECKING,
 )
 
-import six
-from six.moves.queue import Queue
-from wandb.proto import wandb_internal_pb2
-from wandb.proto.wandb_internal_pb2 import Record, Result
+from wandb.proto.wandb_internal_pb2 import (
+    HistoryRecord,
+    MetricRecord,
+    Record,
+    Result,
+    SampledHistoryItem,
+    SummaryItem,
+    SummaryRecord,
+)
 
-from . import meta, sample, stats
-from . import tb_watcher
+from . import meta, sample, stats, tb_watcher
 from .settings_static import SettingsStatic
-from ..interface.interface import BackendSender
-from ..lib import handler_util, proto_util
+from ..interface.interface_queue import InterfaceQueue
+from ..lib import handler_util, proto_util, tracelog
+
+if TYPE_CHECKING:
+    from wandb.proto.wandb_internal_pb2 import (
+        ArtifactDoneRequest,
+        MetricSummary,
+        PartialHistoryRequest,
+    )
+
 
 SummaryDict = Dict[str, Any]
 
@@ -56,21 +65,23 @@ def _dict_nested_set(target: Dict[str, Any], key_list: Sequence[str], v: Any) ->
 class HandleManager(object):
     _consolidated_summary: SummaryDict
     _sampled_history: Dict[str, sample.UniformSampleAccumulator]
+    _partial_history: Dict[str, Any]
     _settings: SettingsStatic
     _record_q: "Queue[Record]"
     _result_q: "Queue[Result]"
     _stopped: Event
     _sender_q: "Queue[Record]"
     _writer_q: "Queue[Record]"
-    _interface: BackendSender
+    _interface: InterfaceQueue
     _system_stats: Optional[stats.SystemStats]
     _tb_watcher: Optional[tb_watcher.TBWatcher]
-    _metric_defines: Dict[str, wandb_internal_pb2.MetricRecord]
-    _metric_globs: Dict[str, wandb_internal_pb2.MetricRecord]
+    _metric_defines: Dict[str, MetricRecord]
+    _metric_globs: Dict[str, MetricRecord]
     _metric_track: Dict[Tuple[str, ...], float]
     _metric_copy: Dict[Tuple[str, ...], Any]
     _track_time: Optional[float]
     _accumulate_time: float
+    _artifact_xid_done: Dict[str, "ArtifactDoneRequest"]
 
     def __init__(
         self,
@@ -80,7 +91,7 @@ class HandleManager(object):
         stopped: Event,
         sender_q: "Queue[Record]",
         writer_q: "Queue[Record]",
-        interface: BackendSender,
+        interface: InterfaceQueue,
     ) -> None:
         self._settings = settings
         self._record_q = record_q
@@ -101,11 +112,15 @@ class HandleManager(object):
         # keep track of summary from key/val updates
         self._consolidated_summary = dict()
         self._sampled_history = dict()
+        self._partial_history = dict()
         self._metric_defines = dict()
         self._metric_globs = dict()
         self._metric_track = dict()
         self._metric_copy = dict()
 
+        # TODO: implement release protocol to clean this up
+        self._artifact_xid_done = dict()
+
     def __len__(self) -> int:
         return self._record_q.qsize()
 
@@ -129,10 +144,16 @@ class HandleManager(object):
 
     def _dispatch_record(self, record: Record, always_send: bool = False) -> None:
         if not self._settings._offline or always_send:
+            tracelog.log_message_queue(record, self._sender_q)
             self._sender_q.put(record)
         if not record.control.local and self._writer_q:
+            tracelog.log_message_queue(record, self._writer_q)
             self._writer_q.put(record)
 
+    def _respond_result(self, result: Result) -> None:
+        tracelog.log_message_queue(result, self._result_q)
+        self._result_q.put(result)
+
     def debounce(self) -> None:
         pass
 
@@ -152,6 +173,8 @@ class HandleManager(object):
                 # shutdown tensorboard workers so we get all metrics flushed
                 self._tb_watcher.finish()
                 self._tb_watcher = None
+        elif state == defer.FLUSH_PARTIAL_HISTORY:
+            self._flush_partial_history()
         elif state == defer.FLUSH_SUM:
             self._save_summary(self._consolidated_summary, flush=True)
 
@@ -183,19 +206,20 @@ class HandleManager(object):
         self._dispatch_record(record)
 
     def _save_summary(self, summary_dict: SummaryDict, flush: bool = False) -> None:
-        summary = wandb_internal_pb2.SummaryRecord()
-        for k, v in six.iteritems(summary_dict):
+        summary = SummaryRecord()
+        for k, v in summary_dict.items():
             update = summary.update.add()
             update.key = k
             update.value_json = json.dumps(v)
-        record = wandb_internal_pb2.Record(summary=summary)
+        record = Record(summary=summary)
         if flush:
             self._dispatch_record(record)
         elif not self._settings._offline:
+            tracelog.log_message_queue(record, self._sender_q)
             self._sender_q.put(record)
 
-    def _save_history(self, record: Record) -> None:
-        for item in record.history.item:
+    def _save_history(self, history: HistoryRecord,) -> None:
+        for item in history.item:
             # TODO(jhr) save nested keys?
             k = item.key
             v = json.loads(item.value_json)
@@ -205,7 +229,7 @@ class HandleManager(object):
 
     def _update_summary_metrics(
         self,
-        s: wandb_internal_pb2.MetricSummary,
+        s: "MetricSummary",
         kl: List[str],
         v: "numbers.Real",
         float_v: float,
@@ -267,10 +291,7 @@ class HandleManager(object):
         return updated
 
     def _update_summary_leaf(
-        self,
-        kl: List[str],
-        v: Any,
-        d: Optional[wandb_internal_pb2.MetricRecord] = None,
+        self, kl: List[str], v: Any, d: Optional[MetricRecord] = None,
     ) -> bool:
         has_summary = d and d.HasField("summary")
         if len(kl) == 1:
@@ -301,17 +322,14 @@ class HandleManager(object):
         return False
 
     def _update_summary_list(
-        self,
-        kl: List[str],
-        v: Any,
-        d: Optional[wandb_internal_pb2.MetricRecord] = None,
+        self, kl: List[str], v: Any, d: Optional[MetricRecord] = None,
     ) -> bool:
         metric_key = ".".join([k.replace(".", "\\.") for k in kl])
         d = self._metric_defines.get(metric_key, d)
         # if the dict has _type key, its a wandb table object
         if isinstance(v, dict) and not handler_util.metric_is_wandb_dict(v):
             updated = False
-            for nk, nv in six.iteritems(v):
+            for nk, nv in v.items():
                 if self._update_summary_list(kl=kl[:] + [nk], v=nv, d=d):
                     updated = True
             return updated
@@ -325,7 +343,7 @@ class HandleManager(object):
 
     def _update_summary_media_objects(self, v: Dict[str, Any]) -> Dict[str, Any]:
         # For now, non recursive - just top level
-        for nk, nv in six.iteritems(v):
+        for nk, nv in v.items():
             if (
                 isinstance(nv, dict)
                 and handler_util.metric_is_wandb_dict(nv)
@@ -344,17 +362,19 @@ class HandleManager(object):
             self._consolidated_summary.update(history_dict)
             return True
         updated = False
-        for k, v in six.iteritems(history_dict):
+        for k, v in history_dict.items():
             if self._update_summary_list(kl=[k], v=v):
                 updated = True
         return updated
 
-    def _history_assign_step(self, record: Record, history_dict: Dict) -> None:
-        has_step = record.history.HasField("step")
-        item = record.history.item.add()
+    def _history_assign_step(
+        self, history: HistoryRecord, history_dict: Dict[str, Any],
+    ) -> None:
+        has_step = history.HasField("step")
+        item = history.item.add()
         item.key = "_step"
         if has_step:
-            step = record.history.step.num
+            step = history.step.num
             history_dict["_step"] = step
             item.value_json = json.dumps(step)
             self._step = step + 1
@@ -363,17 +383,15 @@ class HandleManager(object):
             item.value_json = json.dumps(self._step)
             self._step += 1
 
-    def _history_define_metric(
-        self, hkey: str
-    ) -> Optional[wandb_internal_pb2.MetricRecord]:
+    def _history_define_metric(self, hkey: str) -> Optional[MetricRecord]:
         """check for hkey match in glob metrics, return defined metric."""
         # Dont define metric for internal metrics
         if hkey.startswith("_"):
             return None
-        for k, mglob in six.iteritems(self._metric_globs):
+        for k, mglob in self._metric_globs.items():
             if k.endswith("*"):
                 if hkey.startswith(k[:-1]):
-                    m = wandb_internal_pb2.MetricRecord()
+                    m = MetricRecord()
                     m.CopyFrom(mglob)
                     m.ClearField("glob_name")
                     m.options.defined = False
@@ -382,7 +400,11 @@ class HandleManager(object):
         return None
 
     def _history_update_leaf(
-        self, kl: List[str], v: Any, history_dict: Dict, update_history: Dict[str, Any]
+        self,
+        kl: List[str],
+        v: Any,
+        history_dict: Dict[str, Any],
+        update_history: Dict[str, Any],
     ) -> None:
         hkey = ".".join([k.replace(".", "\\.") for k in kl])
         m = self._metric_defines.get(hkey)
@@ -390,7 +412,7 @@ class HandleManager(object):
             m = self._history_define_metric(hkey)
             if not m:
                 return
-            mr = wandb_internal_pb2.Record()
+            mr = Record()
             mr.metric.CopyFrom(m)
             mr.control.local = True  # Dont store this, just send it
             self._handle_defined_metric(mr)
@@ -403,10 +425,14 @@ class HandleManager(object):
                     update_history[m.step_metric] = step
 
     def _history_update_list(
-        self, kl: List[str], v: Any, history_dict: Dict, update_history: Dict[str, Any]
+        self,
+        kl: List[str],
+        v: Any,
+        history_dict: Dict[str, Any],
+        update_history: Dict[str, Any],
     ) -> None:
         if isinstance(v, dict):
-            for nk, nv in six.iteritems(v):
+            for nk, nv in v.items():
                 self._history_update_list(
                     kl=kl[:] + [nk],
                     v=nv,
@@ -418,21 +444,24 @@ class HandleManager(object):
             kl=kl, v=v, history_dict=history_dict, update_history=update_history
         )
 
-    def _history_update(self, record: Record, history_dict: Dict) -> None:
-        # if syncing an old run, we can skip this logic
+    def _history_update(
+        self, history: HistoryRecord, history_dict: Dict[str, Any],
+    ) -> None:
+
+        #  if syncing an old run, we can skip this logic
         if history_dict.get("_step") is None:
-            self._history_assign_step(record, history_dict)
+            self._history_assign_step(history, history_dict)
 
         update_history: Dict[str, Any] = {}
         # Look for metric matches
         if self._metric_defines or self._metric_globs:
-            for hkey, hval in six.iteritems(history_dict):
+            for hkey, hval in history_dict.items():
                 self._history_update_list([hkey], hval, history_dict, update_history)
 
         if update_history:
             history_dict.update(update_history)
-            for k, v in six.iteritems(update_history):
-                item = record.history.item.add()
+            for k, v in update_history.items():
+                item = history.item.add()
                 item.key = k
                 item.value_json = json.dumps(v)
 
@@ -442,16 +471,38 @@ class HandleManager(object):
         # Inject _runtime if it is not present
         if history_dict is not None:
             if "_runtime" not in history_dict:
-                self._history_assign_runtime(record, history_dict)
+                self._history_assign_runtime(record.history, history_dict)
 
-        self._history_update(record, history_dict)
+        self._history_update(record.history, history_dict)
         self._dispatch_record(record)
-        self._save_history(record)
-
+        self._save_history(record.history)
         updated = self._update_summary(history_dict)
         if updated:
             self._save_summary(self._consolidated_summary)
 
+    def _flush_partial_history(
+        self, partial_history: Optional["PartialHistoryRequest"] = None,
+    ) -> None:
+        if self._partial_history:
+            history = HistoryRecord()
+            for k, v in self._partial_history.items():
+                item = history.item.add()
+                item.key = k
+                item.value_json = json.dumps(v)
+            if partial_history and partial_history.HasField("step"):
+                history.step.MergeFrom(partial_history.step)
+            self.handle_history(Record(history=history))
+            self._partial_history = {}
+
+    def handle_request_partial_history(self, record: Record) -> None:
+        history_dict = proto_util.dict_from_proto_list(
+            record.request.partial_history.item
+        )
+        self._partial_history.update(history_dict)
+
+        if record.request.partial_history.action.flush:
+            self._flush_partial_history(record.request.partial_history)
+
     def handle_summary(self, record: Record) -> None:
         summary = record.summary
         for item in summary.update:
@@ -515,9 +566,50 @@ class HandleManager(object):
     def handle_request_check_version(self, record: Record) -> None:
         self._dispatch_record(record)
 
+    def handle_request_attach(self, record: Record) -> None:
+        self._dispatch_record(record)
+
     def handle_request_log_artifact(self, record: Record) -> None:
         self._dispatch_record(record)
 
+    def handle_request_artifact_send(self, record: Record) -> None:
+        assert record.control.req_resp
+        result = proto_util._result_from_record(record)
+
+        self._dispatch_record(record)
+
+        # send response immediately, the request will be polled for result
+        xid = record.uuid
+        result.response.artifact_send_response.xid = xid
+        self._respond_result(result)
+
+    def handle_request_artifact_poll(self, record: Record) -> None:
+        assert record.control.req_resp
+        xid = record.request.artifact_poll.xid
+        assert xid
+
+        result = proto_util._result_from_record(record)
+        done_req = self._artifact_xid_done.get(xid)
+        if done_req:
+            result.response.artifact_poll_response.artifact_id = done_req.artifact_id
+            result.response.artifact_poll_response.error_message = (
+                done_req.error_message
+            )
+            result.response.artifact_poll_response.ready = True
+        self._respond_result(result)
+
+    def handle_request_artifact_done(self, record: Record) -> None:
+        assert not record.control.req_resp
+        done_req = record.request.artifact_done
+        xid = done_req.xid
+        assert xid
+
+        self._artifact_xid_done[xid] = done_req
+
+    # def handle_request_artifact_release(self, record: Record) -> None:
+    #     assert record.control.req_resp
+    #     # TODO: implement release protocol to clean up _artifact_xid_done dict
+
     def handle_telemetry(self, record: Record) -> None:
         self._dispatch_record(record)
 
@@ -550,8 +642,8 @@ class HandleManager(object):
 
         if run_start.run.resumed:
             self._step = run_start.run.starting_step
-        result = wandb_internal_pb2.Result(uuid=record.uuid)
-        self._result_q.put(result)
+        result = proto_util._result_from_record(record)
+        self._respond_result(result)
 
     def handle_request_resume(self, record: Record) -> None:
         if self._system_stats is not None:
@@ -579,14 +671,17 @@ class HandleManager(object):
     def handle_request_network_status(self, record: Record) -> None:
         self._dispatch_record(record)
 
+    def handle_request_status(self, record: Record) -> None:
+        self._dispatch_record(record, always_send=True)
+
     def handle_request_get_summary(self, record: Record) -> None:
-        result = wandb_internal_pb2.Result(uuid=record.uuid)
-        for key, value in six.iteritems(self._consolidated_summary):
-            item = wandb_internal_pb2.SummaryItem()
+        result = proto_util._result_from_record(record)
+        for key, value in self._consolidated_summary.items():
+            item = SummaryItem()
             item.key = key
             item.value_json = json.dumps(value)
             result.response.get_summary_response.item.append(item)
-        self._result_q.put(result)
+        self._respond_result(result)
 
     def handle_tbrecord(self, record: Record) -> None:
         logger.info("handling tbrecord: %s", record)
@@ -595,40 +690,40 @@ class HandleManager(object):
             self._tb_watcher.add(tbrecord.log_dir, tbrecord.save, tbrecord.root_dir)
         self._dispatch_record(record)
 
-    def _handle_defined_metric(self, record: wandb_internal_pb2.Record) -> None:
+    def _handle_defined_metric(self, record: Record) -> None:
         metric = record.metric
         if metric._control.overwrite:
-            self._metric_defines.setdefault(
-                metric.name, wandb_internal_pb2.MetricRecord()
-            ).CopyFrom(metric)
+            self._metric_defines.setdefault(metric.name, MetricRecord()).CopyFrom(
+                metric
+            )
         else:
-            self._metric_defines.setdefault(
-                metric.name, wandb_internal_pb2.MetricRecord()
-            ).MergeFrom(metric)
+            self._metric_defines.setdefault(metric.name, MetricRecord()).MergeFrom(
+                metric
+            )
 
         # before dispatching, make sure step_metric is defined, if not define it and
         # dispatch it locally first
         metric = self._metric_defines[metric.name]
         if metric.step_metric and metric.step_metric not in self._metric_defines:
-            m = wandb_internal_pb2.MetricRecord(name=metric.step_metric)
+            m = MetricRecord(name=metric.step_metric)
             self._metric_defines[metric.step_metric] = m
-            mr = wandb_internal_pb2.Record()
+            mr = Record()
             mr.metric.CopyFrom(m)
             mr.control.local = True  # Dont store this, just send it
             self._dispatch_record(mr)
 
         self._dispatch_record(record)
 
-    def _handle_glob_metric(self, record: wandb_internal_pb2.Record) -> None:
+    def _handle_glob_metric(self, record: Record) -> None:
         metric = record.metric
         if metric._control.overwrite:
-            self._metric_globs.setdefault(
-                metric.glob_name, wandb_internal_pb2.MetricRecord()
-            ).CopyFrom(metric)
+            self._metric_globs.setdefault(metric.glob_name, MetricRecord()).CopyFrom(
+                metric
+            )
         else:
-            self._metric_globs.setdefault(
-                metric.glob_name, wandb_internal_pb2.MetricRecord()
-            ).MergeFrom(metric)
+            self._metric_globs.setdefault(metric.glob_name, MetricRecord()).MergeFrom(
+                metric
+            )
         self._dispatch_record(record)
 
     def handle_metric(self, record: Record) -> None:
@@ -661,9 +756,9 @@ class HandleManager(object):
             self._handle_glob_metric(record)
 
     def handle_request_sampled_history(self, record: Record) -> None:
-        result = wandb_internal_pb2.Result(uuid=record.uuid)
-        for key, sampled in six.iteritems(self._sampled_history):
-            item = wandb_internal_pb2.SampledHistoryItem()
+        result = proto_util._result_from_record(record)
+        for key, sampled in self._sampled_history.items():
+            item = SampledHistoryItem()
             item.key = key
             values: Iterable[Any] = sampled.get()
             if all(isinstance(i, numbers.Integral) for i in values):
@@ -671,12 +766,12 @@ class HandleManager(object):
             elif all(isinstance(i, numbers.Real) for i in values):
                 item.values_float.extend(values)
             result.response.sampled_history_response.item.append(item)
-        self._result_q.put(result)
+        self._respond_result(result)
 
     def handle_request_shutdown(self, record: Record) -> None:
         # TODO(jhr): should we drain things and stop new requests from coming in?
-        result = wandb_internal_pb2.Result(uuid=record.uuid)
-        self._result_q.put(result)
+        result = proto_util._result_from_record(record)
+        self._respond_result(result)
         self._stopped.set()
 
     def finish(self) -> None:
@@ -689,7 +784,9 @@ class HandleManager(object):
 
     next = __next__
 
-    def _history_assign_runtime(self, record: Record, history_dict: Dict) -> None:
+    def _history_assign_runtime(
+        self, history: HistoryRecord, history_dict: Dict[str, Any],
+    ) -> None:
         # _runtime calculation is meaningless if there is no _timestamp
         if "_timestamp" not in history_dict:
             return
@@ -700,6 +797,6 @@ class HandleManager(object):
         history_dict["_runtime"] = int(
             history_dict["_timestamp"] - self._run_start_time
         )
-        item = record.history.item.add()
+        item = history.item.add()
         item.key = "_runtime"
         item.value_json = json.dumps(history_dict[item.key])
