@@ -1,27 +1,25 @@
-#
-from gql import Client, gql  # type: ignore
-from gql.client import RetryError  # type: ignore
-from gql.transport.requests import RequestsHTTPTransport  # type: ignore
-import datetime
+from wandb_gql import Client, gql  # type: ignore
+from wandb_gql.client import RetryError  # type: ignore
+from wandb_gql.transport.requests import RequestsHTTPTransport  # type: ignore
+
 import ast
+import base64
+from copy import deepcopy
+import datetime
+from io import BytesIO
+import json
 import os
 from pkg_resources import parse_version  # type: ignore
-import json
-import yaml
 import re
-import click
-import logging
 import requests
+import logging
+import socket
 import sys
 
-if os.name == "posix" and sys.version_info[0] < 3:
-    import subprocess32 as subprocess  # type: ignore
-else:
-    import subprocess  # type: ignore[no-redef]
-
-from copy import deepcopy
+import click
 import six
-from six import BytesIO
+import yaml
+
 import wandb
 from wandb import __version__
 from wandb import env
@@ -39,13 +37,13 @@ from .progress import Progress
 logger = logging.getLogger(__name__)
 
 
-class Api(object):
+class Api:
     """W&B Internal Api wrapper
 
     Note:
         Settings are automatically overridden by looking for
-        a `wandb/settings` file in the current working directory or it's parent
-        directory.  If none can be found, we look in the current users home
+        a `wandb/settings` file in the current working directory or its parent
+        directory. If none can be found, we look in the current user's home
         directory.
 
     Arguments:
@@ -96,7 +94,7 @@ class Api(object):
                 # https://bugs.python.org/issue22889
                 timeout=self.HTTP_TIMEOUT,
                 auth=("api", self.api_key or ""),
-                url="%s/graphql" % self.settings("base_url"),
+                url=f"{self.settings('base_url')}/graphql",
             )
         )
         self.retry_callback = retry_callback
@@ -115,8 +113,18 @@ class Api(object):
             retry.retriable(retry_timedelta=retry_timedelta)(self.upload_file)
         )
         self._client_id_mapping = {}
+        # Large file uploads to azure can optionally use their SDK
+        self._azure_blob_module = util.get_module("azure.storage.blob")
 
-        self.query_types, self.server_info_types = None, None
+        (
+            self.query_types,
+            self.server_info_types,
+            self.server_use_artifact_input_info,
+        ) = (
+            None,
+            None,
+            None,
+        )
 
     def reauth(self):
         """Ensures the current api key is set in the transport"""
@@ -146,7 +154,7 @@ class Api(object):
         if "errors" in data and isinstance(data["errors"], list):
             for err in data["errors"]:
                 # Our tests and potentially some api endpoints return a string error?
-                if isinstance(err, six.string_types):
+                if isinstance(err, str):
                     err = {"message": err}
                 if not err.get("message"):
                     continue
@@ -282,7 +290,6 @@ class Api(object):
 
     @normalize_exceptions
     def server_info_introspection(self):
-
         query_string = """
            query ProbeServerCapabilities {
                QueryType: __type(name: "Query") {
@@ -314,6 +321,44 @@ class Api(object):
             ]
         return (self.query_types, self.server_info_types)
 
+    def server_use_artifact_input_introspection(self):
+        query_string = """
+           query ProbeServerUseArtifactInput {
+               UseArtifactInputInfoType: __type(name: "UseArtifactInput") {
+                   name
+                   inputFields {
+                       name
+                   }
+                }
+            }
+        """
+
+        if self.server_use_artifact_input_info is None:
+            query = gql(query_string)
+            res = self.gql(query)
+            self.server_use_artifact_input_info = [
+                field.get("name", "")
+                for field in res.get("UseArtifactInputInfoType", {}).get(
+                    "inputFields", [{}]
+                )
+            ]
+        return self.server_use_artifact_input_info
+
+    @normalize_exceptions
+    def launch_agent_introspection(self):
+        query = gql(
+            """
+            query LaunchAgentIntrospection {
+                LaunchAgentType: __type(name: "LaunchAgent") {
+                    name
+                }
+            }
+        """
+        )
+
+        res = self.gql(query)
+        return res.get("LaunchAgentType") or None
+
     @normalize_exceptions
     def viewer(self):
         query = gql(
@@ -402,7 +447,7 @@ class Api(object):
         """
         query = gql(
             """
-        query Models($entity: String!) {
+        query EntityProjects($entity: String) {
             models(first: 10, entityName: $entity) {
                 edges {
                     node {
@@ -434,7 +479,7 @@ class Api(object):
         """
         query = gql(
             """
-        query Models($entity: String, $project: String!) {
+        query ProjectDetails($entity: String, $project: String) {
             model(name: $project, entityName: $entity) {
                 id
                 name
@@ -464,7 +509,7 @@ class Api(object):
         """
         query = gql(
             """
-        query Sweep($entity: String, $project: String, $sweep: String!, $specs: [JSONString!]!) {
+        query SweepWithRuns($entity: String, $project: String, $sweep: String!, $specs: [JSONString!]!) {
             project(name: $project, entityName: $entity) {
                 sweep(sweepName: $sweep) {
                     id
@@ -533,7 +578,7 @@ class Api(object):
         """
         query = gql(
             """
-        query Buckets($model: String!, $entity: String!) {
+        query ProjectRuns($model: String!, $entity: String) {
             model(name: $model, entityName: $entity) {
                 buckets(first: 10) {
                     edges {
@@ -624,9 +669,9 @@ class Api(object):
         """
         query = gql(
             """
-        query Model(
+        query RunConfigs(
             $name: String!,
-            $entity: String!,
+            $entity: String,
             $run: String!,
             $pattern: String!,
             $includeConfig: Boolean!,
@@ -709,7 +754,7 @@ class Api(object):
         """
         query = gql(
             """
-        query Model($project: String!, $entity: String, $name: String!) {
+        query RunResumeStatus($project: String, $entity: String, $name: String!) {
             model(name: $project, entityName: $entity) {
                 id
                 name
@@ -754,7 +799,7 @@ class Api(object):
     def check_stop_requested(self, project_name, entity_name, run_id):
         query = gql(
             """
-        query Model($projectName: String, $entityName: String, $runId: String!) {
+        query RunStoppedStatus($projectName: String, $entityName: String, $runId: String!) {
             project(name:$projectName, entityName:$entityName) {
                 run(name:$runId) {
                     stopped
@@ -823,7 +868,7 @@ class Api(object):
     def get_project_run_queues(self, entity, project):
         query = gql(
             """
-        query Project($entity: String!, $projectName: String!){
+        query ProjectRunQueues($entity: String!, $projectName: String!){
             project(entityName: $entity, name: $projectName) {
                 runQueues {
                     id
@@ -866,7 +911,6 @@ class Api(object):
                 success
                 queueID
             }
-            
         }
         """
         )
@@ -953,11 +997,16 @@ class Api(object):
         return response["pushToRunQueue"]
 
     @normalize_exceptions
-    def pop_from_run_queue(self, queue_name, entity=None, project=None):
+    def pop_from_run_queue(self, queue_name, entity=None, project=None, agent_id=None):
         mutation = gql(
             """
-        mutation popFromRunQueue($entity: String!, $project: String!, $queueName: String!)  {
-            popFromRunQueue(input: { entityName: $entity, projectName: $project, queueName: $queueName }) {
+        mutation popFromRunQueue($entity: String!, $project: String!, $queueName: String!, $launchAgentId: ID)  {
+            popFromRunQueue(input: {
+                entityName: $entity,
+                projectName: $project,
+                queueName: $queueName,
+                launchAgentId: $launchAgentId
+            }) {
                 runQueueItemId
                 runSpec
             }
@@ -970,6 +1019,7 @@ class Api(object):
                 "entity": entity,
                 "project": project,
                 "queueName": queue_name,
+                "launchAgentId": agent_id,
             },
         )
         return response["popFromRunQueue"]
@@ -994,6 +1044,119 @@ class Api(object):
             )
         return response["ackRunQueueItem"]["success"]
 
+    @normalize_exceptions
+    def create_launch_agent(self, entity, project, queues, gorilla_agent_support):
+        project_queues = self.get_project_run_queues(entity, project)
+        if not project_queues:
+            # create default queue if it doesn't already exist
+            default = self.create_run_queue(
+                entity, project, "default", access="PROJECT"
+            )
+            if default is None or default.get("queueID") is None:
+                raise CommError(
+                    "Unable to create default queue for {}/{}. No queues for agent to poll".format(
+                        entity, project
+                    )
+                )
+            project_queues = [{"id": default["queueID"], "name": "default"}]
+        polling_queue_ids = [
+            q["id"] for q in project_queues if q["name"] in queues
+        ]  # filter to poll specified queues
+        if len(polling_queue_ids) != len(queues):
+            raise CommError(
+                "Could not start launch agent: Not all of requested queues ({}) found. Available queues for this project: {}".format(
+                    ", ".join(queues), ",".join([q["name"] for q in project_queues])
+                )
+            )
+
+        if not gorilla_agent_support:
+            # if gorilla doesn't support launch agents, return a client-generated id
+            return {
+                "success": True,
+                "launchAgentId": None,
+            }
+
+        hostname = socket.gethostname()
+        mutation = gql(
+            """
+            mutation createLaunchAgent($entity: String!, $project: String!, $queues: [ID!]!, $hostname: String!){
+                createLaunchAgent(
+                    input: {
+                        entityName: $entity,
+                        projectName: $project,
+                        runQueues: $queues,
+                        hostname: $hostname
+                    }
+                ) {
+                    launchAgentId
+                }
+            }
+            """
+        )
+        variable_values = {
+            "entity": entity,
+            "project": project,
+            "queues": polling_queue_ids,
+            "hostname": hostname,
+        }
+        return self.gql(mutation, variable_values)["createLaunchAgent"]
+
+    @normalize_exceptions
+    def update_launch_agent_status(self, agent_id, status, gorilla_agent_support):
+        if not gorilla_agent_support:
+            # if gorilla doesn't support launch agents, this is a no-op
+            return {
+                "success": True,
+            }
+
+        mutation = gql(
+            """
+            mutation updateLaunchAgent($agentId: ID!, $agentStatus: String){
+                updateLaunchAgent(
+                    input: {
+                        launchAgentId: $agentId
+                        agentStatus: $agentStatus
+                    }
+                ) {
+                    success
+                }
+            }
+            """
+        )
+        variable_values = {
+            "agentId": agent_id,
+            "agentStatus": status,
+        }
+        return self.gql(mutation, variable_values)["updateLaunchAgent"]
+
+    @normalize_exceptions
+    def get_launch_agent(self, agent_id, gorilla_agent_support):
+        if not gorilla_agent_support:
+            return {
+                "id": None,
+                "name": "",
+                "stopPolling": False,
+            }
+        query = gql(
+            """
+            query LaunchAgent($agentId: ID!) {
+                launchAgent(id: $agentId) {
+                    id
+                    name
+                    runQueues
+                    hostname
+                    agentStatus
+                    stopPolling
+                    heartbeatAt
+                }
+            }
+            """
+        )
+        variable_values = {
+            "agentId": agent_id,
+        }
+        return self.gql(query, variable_values)["launchAgent"]
+
     @normalize_exceptions
     def upsert_run(
         self,
@@ -1146,7 +1309,7 @@ class Api(object):
     def get_run_info(self, entity, project, name):
         query = gql(
             """
-        query Run($project: String!, $entity: String!, $name: String!) {
+        query RunInfo($project: String!, $entity: String!, $name: String!) {
             project(name: $project, entityName: $entity) {
                 run(name: $name) {
                     runInfo {
@@ -1207,7 +1370,7 @@ class Api(object):
         """
         query = gql(
             """
-        query Model($name: String!, $files: [String]!, $entity: String!, $run: String!, $description: String) {
+        query RunUploadUrls($name: String!, $files: [String]!, $entity: String, $run: String!, $description: String) {
             model(name: $name, entityName: $entity) {
                 bucket(name: $run, desc: $description) {
                     id
@@ -1268,7 +1431,7 @@ class Api(object):
         """
         query = gql(
             """
-        query Model($name: String!, $entity: String!, $run: String!)  {
+        query RunDownloadUrls($name: String!, $entity: String, $run: String!)  {
             model(name: $name, entityName: $entity) {
                 bucket(name: $run) {
                     files {
@@ -1315,7 +1478,7 @@ class Api(object):
         """
         query = gql(
             """
-        query Model($name: String!, $fileName: String!, $entity: String!, $run: String!)  {
+        query RunDownloadUrl($name: String!, $fileName: String!, $entity: String, $run: String!)  {
             model(name: $name, entityName: $entity) {
                 bucket(name: $run) {
                     files(names: [$fileName]) {
@@ -1387,6 +1550,38 @@ class Api(object):
 
         return path, response
 
+    def upload_file_azure(self, url, file, extra_headers):
+        from azure.core.exceptions import AzureError
+
+        # Configure the client without retries so our existing logic can handle them
+        client = self._azure_blob_module.BlobClient.from_blob_url(
+            url, retry_policy=self._azure_blob_module.LinearRetry(retry_total=0)
+        )
+        try:
+            if extra_headers.get("Content-MD5") is not None:
+                md5 = base64.b64decode(extra_headers["Content-MD5"])
+            else:
+                md5 = None
+            content_settings = self._azure_blob_module.ContentSettings(
+                content_md5=md5, content_type=extra_headers.get("Content-Type"),
+            )
+            client.upload_blob(
+                file,
+                max_concurrency=4,
+                length=len(file),
+                overwrite=True,
+                content_settings=content_settings,
+            )
+        except AzureError as e:
+            if hasattr(e, "response"):
+                response = requests.model.Response()
+                response.status_code = e.response.status_code
+                response.headers = e.response.headers
+                response.raw = e.response.internal_response
+                raise requests.exceptions.RequestException(e.message, response=response)
+            else:
+                raise requests.exceptions.ConnectionError(e.message)
+
     def upload_file(self, url, file, callback=None, extra_headers={}):
         """Uploads a file to W&B with failure resumption
 
@@ -1403,11 +1598,23 @@ class Api(object):
         response = None
         progress = Progress(file, callback=callback)
         try:
-            response = requests.put(url, data=progress, headers=extra_headers)
-            response.raise_for_status()
+            if "x-ms-blob-type" in extra_headers and self._azure_blob_module:
+                response = self.upload_file_azure(url, progress, extra_headers)
+            else:
+                if "x-ms-blob-type" in extra_headers:
+                    wandb.termwarn(
+                        "Azure uploads over 256MB require the azure SDK, install with pip install wandb[azure]",
+                        repeat=False,
+                    )
+                response = requests.put(url, data=progress, headers=extra_headers)
+                response.raise_for_status()
         except requests.exceptions.RequestException as e:
-            logger.error("upload_file exception {} {}".format(url, e))
-            status_code = e.response.status_code if e.response != None else 0
+            logger.error("upload_file exception {}: {}".format(url, e))
+            request_headers = e.request.headers if e.request is not None else ""
+            logger.error("upload_file request headers: {}".format(request_headers))
+            response_content = e.response.content if e.response is not None else ""
+            logger.error("upload_file response body: {}".format(response_content))
+            status_code = e.response.status_code if e.response is not None else 0
             # We need to rewind the file for the next retry (the file passed in is seeked to 0)
             progress.rewind()
             # Retry errors from cloud storage or local network issues
@@ -1435,8 +1642,8 @@ class Api(object):
             """
         mutation CreateAgent(
             $host: String!
-            $projectName: String!,
-            $entityName: String!,
+            $projectName: String,
+            $entityName: String,
             $sweep: String!
         ) {
             createAgent(input: {
@@ -1511,6 +1718,10 @@ class Api(object):
         }
         """
         )
+
+        if agent_id is None:
+            raise ValueError("Cannot call heartbeat with an unregistered agent.")
+
         try:
             response = self.gql(
                 mutation,
@@ -1803,7 +2014,7 @@ class Api(object):
                     else open(normal_name, "rb")
                 )
             except IOError:
-                print("%s does not exist" % file_name)
+                print(f"{file_name} does not exist")
                 continue
             if progress:
                 if hasattr(progress, "__call__"):
@@ -1838,21 +2049,27 @@ class Api(object):
         return responses
 
     def use_artifact(
-        self, artifact_id, entity_name=None, project_name=None, run_name=None
+        self,
+        artifact_id,
+        entity_name=None,
+        project_name=None,
+        run_name=None,
+        use_as=None,
     ):
-        query = gql(
-            """
+        query_template = """
         mutation UseArtifact(
             $entityName: String!,
             $projectName: String!,
             $runName: String!,
-            $artifactID: ID!
+            $artifactID: ID!,
+            _USED_AS_TYPE_
         ) {
             useArtifact(input: {
                 entityName: $entityName,
                 projectName: $projectName,
                 runName: $runName,
-                artifactID: $artifactID
+                artifactID: $artifactID,
+                _USED_AS_VALUE_
             }) {
                 artifact {
                     id
@@ -1866,7 +2083,19 @@ class Api(object):
             }
         }
         """
-        )
+
+        artifact_types = self.server_use_artifact_input_introspection()
+        if "usedAs" in artifact_types:
+            query_template = query_template.replace(
+                "_USED_AS_TYPE_", "$usedAs: String"
+            ).replace("_USED_AS_VALUE_", "usedAs: $usedAs")
+        else:
+            query_template = query_template.replace("_USED_AS_TYPE_", "").replace(
+                "_USED_AS_VALUE_", ""
+            )
+
+        query = gql(query_template)
+
         entity_name = entity_name or self.settings("entity")
         project_name = project_name or self.settings("project")
         run_name = run_name or self.current_run_id
@@ -1878,6 +2107,7 @@ class Api(object):
                 "projectName": project_name,
                 "runName": run_name,
                 "artifactID": artifact_id,
+                "usedAs": use_as,
             },
         )
 
@@ -1938,8 +2168,8 @@ class Api(object):
         aliases=None,
         distributed_id=None,
         is_user_created=False,
+        enable_digest_deduplication=False,
     ):
-        # TODO: Ignore clientID and sequenceClientID if server can't handle it
         _, server_info = self.viewer_server_info()
         max_cli_version = server_info.get("cliVersionInfo", {}).get(
             "max_cli_version", None
@@ -1947,6 +2177,9 @@ class Api(object):
         can_handle_client_id = max_cli_version is None or parse_version(
             "0.11.0"
         ) <= parse_version(max_cli_version)
+        can_handle_dedupe = max_cli_version is None or parse_version(
+            "0.12.10"
+        ) <= parse_version(max_cli_version)
 
         mutation = gql(
             """
@@ -1964,6 +2197,7 @@ class Api(object):
             %s
             %s
             %s
+            %s
         ) {
             createArtifact(input: {
                 artifactTypeName: $artifactTypeName,
@@ -1980,6 +2214,7 @@ class Api(object):
                 %s
                 %s
                 %s
+                %s
             }) {
                 artifact {
                     id
@@ -2002,14 +2237,18 @@ class Api(object):
         """
             %
             # For backwards compatibility with older backends that don't support
-            # distributed writers.
+            # distributed writers or digest deduplication.
             (
                 "$distributedID: String," if distributed_id else "",
                 "$clientID: ID!," if can_handle_client_id else "",
                 "$sequenceClientID: ID!," if can_handle_client_id else "",
+                "$enableDigestDeduplication: Boolean," if can_handle_dedupe else "",
                 "distributedID: $distributedID," if distributed_id else "",
                 "clientID: $clientID," if can_handle_client_id else "",
                 "sequenceClientID: $sequenceClientID," if can_handle_client_id else "",
+                "enableDigestDeduplication: $enableDigestDeduplication,"
+                if can_handle_dedupe
+                else "",
             )
         )
 
@@ -2040,6 +2279,7 @@ class Api(object):
                 if metadata
                 else None,
                 "distributedID": distributed_id,
+                "enableDigestDeduplication": enable_digest_deduplication,
             },
         )
         av = response["createArtifact"]["artifact"]
