@@ -13,6 +13,7 @@ import logging
 import os
 import time
 from typing import Any, Dict, Generator, List, NewType, Optional, Tuple
+from typing import cast, TYPE_CHECKING
 
 from pkg_resources import parse_version
 import requests
@@ -22,17 +23,29 @@ import wandb
 from wandb import util
 from wandb.filesync.dir_watcher import DirWatcher
 from wandb.proto import wandb_internal_pb2
-from wandb.proto.wandb_internal_pb2 import HttpResponse
 
 from . import artifacts
 from . import file_stream
 from . import internal_api
-from . import settings_static
 from . import update
 from .file_pusher import FilePusher
+from .settings_static import SettingsDict, SettingsStatic
 from ..interface import interface
+from ..interface.interface_queue import InterfaceQueue
 from ..lib import config_util, filenames, proto_util, telemetry
-from ..lib.git import GitRepo
+from ..lib import tracelog
+from ..lib.proto_util import message_to_dict
+
+if TYPE_CHECKING:
+    from wandb.proto.wandb_internal_pb2 import (
+        ArtifactRecord,
+        HttpResponse,
+        LocalInfo,
+        Record,
+        Result,
+        RunExitResult,
+        RunRecord,
+    )
 
 
 logger = logging.getLogger(__name__)
@@ -58,13 +71,62 @@ def _framework_priority(
     yield imp.sklearn, "sklearn"
 
 
-class SendManager(object):
+class ResumeState:
+    resumed: bool
+    step: int
+    history: int
+    events: int
+    output: int
+    runtime: int
+    wandb_runtime: Optional[int]
+    summary: Optional[Dict[str, Any]]
+    config: Optional[Dict[str, Any]]
+
+    def __init__(self) -> None:
+        self.resumed = False
+        self.step = 0
+        self.history = 0
+        self.events = 0
+        self.output = 0
+        self.runtime = 0
+        # wandb_runtime is the canonical runtime (stored in summary._wandb.runtime)
+        self.wandb_runtime = None
+        self.summary = None
+        self.config = None
+
+    def __str__(self) -> str:
+        obj = ",".join(map(lambda it: f"{it[0]}={it[1]}", vars(self).items()))
+        return f"ResumeState({obj})"
+
+
+class SendManager:
+
+    _settings: SettingsStatic
+    _record_q: "Queue[Record]"
+    _result_q: "Queue[Result]"
+    _interface: InterfaceQueue
+    _api_settings: Dict[str, str]
+    _partial_output: Dict[str, str]
 
     _telemetry_obj: telemetry.TelemetryRecord
+    _fs: "Optional[file_stream.FileStreamApi]"
+    _run: "Optional[RunRecord]"
+    _entity: "Optional[str]"
+    _project: "Optional[str]"
+    _dir_watcher: "Optional[DirWatcher]"
+    _pusher: "Optional[FilePusher]"
+    _exit_result: "Optional[RunExitResult]"
+    _resume_state: ResumeState
+    _cached_server_info: Dict[str, Any]
+    _cached_viewer: Dict[str, Any]
 
     def __init__(
-        self, settings, record_q, result_q, interface,
-    ):
+        self,
+        settings: SettingsStatic,
+        record_q: "Queue[Record]",
+        result_q: "Queue[Result]",
+        interface: InterfaceQueue,
+    ) -> None:
         self._settings = settings
         self._record_q = record_q
         self._result_q = result_q
@@ -83,7 +145,7 @@ class SendManager(object):
         self._project = None
 
         # keep track of config from key/val updates
-        self._consolidated_config: DictNoValues = dict()
+        self._consolidated_config: DictNoValues = cast(DictNoValues, dict())
         self._start_time: int = 0
         self._telemetry_obj = telemetry.TelemetryRecord()
         self._config_metric_pbdict_list: List[Dict[int, Any]] = []
@@ -92,23 +154,11 @@ class SendManager(object):
         self._config_metric_index_dict: Dict[str, int] = {}
         self._config_metric_dict: Dict[str, wandb_internal_pb2.MetricRecord] = {}
 
-        self._cached_server_info: Dict[str, Any] = dict()
-        self._cached_viewer: Dict[str, Any] = dict()
+        self._cached_server_info = dict()
+        self._cached_viewer = dict()
 
         # State updated by resuming
-        self._resume_state = {
-            "step": 0,
-            "history": 0,
-            "events": 0,
-            "output": 0,
-            "runtime": 0,
-            "summary": None,
-            "config": None,
-            "resumed": False,
-        }
-
-        # State added when run_exit needs results
-        self._exit_sync_uuid = None
+        self._resume_state = ResumeState()
 
         # State added when run_exit is complete
         self._exit_result = None
@@ -130,12 +180,12 @@ class SendManager(object):
         self._exit_code = 0
 
     @classmethod
-    def setup(cls, root_dir):
+    def setup(cls, root_dir: str) -> "SendManager":
         """This is a helper class method to setup a standalone SendManager.
         Currently we're using this primarily for `sync.py`.
         """
         files_dir = os.path.join(root_dir, "files")
-        sd = dict(
+        sd: SettingsDict = dict(
             files_dir=files_dir,
             root_dir=root_dir,
             _start_time=0,
@@ -155,10 +205,10 @@ class SendManager(object):
             email=None,
             silent=None,
         )
-        settings = settings_static.SettingsStatic(sd)
-        record_q = queue.Queue()
-        result_q = queue.Queue()
-        publish_interface = interface.BackendSender(record_q=record_q)
+        settings = SettingsStatic(sd)
+        record_q: "Queue[Record]" = queue.Queue()
+        result_q: "Queue[Result]" = queue.Queue()
+        publish_interface = InterfaceQueue(record_q=record_q)
         return SendManager(
             settings=settings,
             record_q=record_q,
@@ -166,16 +216,16 @@ class SendManager(object):
             interface=publish_interface,
         )
 
-    def __len__(self):
+    def __len__(self) -> int:
         return self._record_q.qsize()
 
-    def retry_callback(self, status, response_text):
+    def retry_callback(self, status: int, response_text: str) -> None:
         response = wandb_internal_pb2.HttpResponse()
         response.http_status_code = status
         response.http_response_text = response_text
         self._retry_q.put(response)
 
-    def send(self, record):
+    def send(self, record: "Record") -> None:
         record_type = record.WhichOneof("record_type")
         assert record_type
         handler_str = "send_" + record_type
@@ -186,11 +236,11 @@ class SendManager(object):
         assert send_handler, "unknown send handler: {}".format(handler_str)
         send_handler(record)
 
-    def send_preempting(self, record):
+    def send_preempting(self, record: "Record") -> None:
         if self._fs:
             self._fs.enqueue_preempting()
 
-    def send_request(self, record):
+    def send_request(self, record: "Record") -> None:
         request_type = record.request.WhichOneof("request_type")
         assert request_type
         handler_str = "send_request_" + request_type
@@ -200,7 +250,11 @@ class SendManager(object):
         assert send_handler, "unknown handle: {}".format(handler_str)
         send_handler(record)
 
-    def _flatten(self, dictionary):
+    def _respond_result(self, result: "Result") -> None:
+        tracelog.log_message_queue(result, self._result_q)
+        self._result_q.put(result)
+
+    def _flatten(self, dictionary: Dict) -> None:
         if type(dictionary) == dict:
             for k, v in list(dictionary.items()):
                 if type(v) == dict:
@@ -209,9 +263,9 @@ class SendManager(object):
                     for k2, v2 in v.items():
                         dictionary[k + "." + k2] = v2
 
-    def send_request_check_version(self, record):
+    def send_request_check_version(self, record: "Record") -> None:
         assert record.control.req_resp
-        result = wandb_internal_pb2.Result(uuid=record.uuid)
+        result = proto_util._result_from_record(record)
         current_version = (
             record.request.check_version.current_version or wandb.__version__
         )
@@ -226,28 +280,46 @@ class SendManager(object):
             delete_message = messages.get("delete_message")
             if delete_message:
                 result.response.check_version_response.delete_message = delete_message
-        self._result_q.put(result)
+        self._respond_result(result)
 
-    def send_request_stop_status(self, record):
+    def _send_request_attach(
+        self,
+        req: wandb_internal_pb2.AttachRequest,
+        resp: wandb_internal_pb2.AttachResponse,
+    ) -> None:
+        attach_id = req.attach_id
+        assert attach_id
+        assert self._run
+        resp.run.CopyFrom(self._run)
+
+    def send_request_attach(self, record: "Record") -> None:
+        assert record.control.req_resp
+        result = proto_util._result_from_record(record)
+        self._send_request_attach(
+            record.request.attach, result.response.attach_response
+        )
+        self._respond_result(result)
+
+    def send_request_stop_status(self, record: "Record") -> None:
         assert record.control.req_resp
 
-        result = wandb_internal_pb2.Result(uuid=record.uuid)
+        result = proto_util._result_from_record(record)
         status_resp = result.response.stop_status_response
         status_resp.run_should_stop = False
-        if self._entity and self._project and self._run.run_id:
+        if self._entity and self._project and self._run and self._run.run_id:
             try:
                 status_resp.run_should_stop = self._api.check_stop_requested(
                     self._project, self._entity, self._run.run_id
                 )
             except Exception as e:
                 logger.warning("Failed to check stop requested status: %s", e)
-        self._result_q.put(result)
+        self._respond_result(result)
 
     def debounce(self) -> None:
         if self._config_needs_debounce:
             self._debounce_config()
 
-    def _debounce_config(self):
+    def _debounce_config(self) -> None:
         config_value_dict = self._config_format(self._consolidated_config)
         # TODO(jhr): check result of upsert_run?
         if self._run:
@@ -257,10 +329,15 @@ class SendManager(object):
         self._config_save(config_value_dict)
         self._config_needs_debounce = False
 
-    def send_request_network_status(self, record):
+    def send_request_status(self, record: "Record") -> None:
+        assert record.control.req_resp
+        result = proto_util._result_from_record(record)
+        self._respond_result(result)
+
+    def send_request_network_status(self, record: "Record") -> None:
         assert record.control.req_resp
 
-        result = wandb_internal_pb2.Result(uuid=record.uuid)
+        result = proto_util._result_from_record(record)
         status_resp = result.response.network_status_response
         while True:
             try:
@@ -269,9 +346,9 @@ class SendManager(object):
                 break
             except Exception as e:
                 logger.warning("Error emptying retry queue: {}".format(e))
-        self._result_q.put(result)
+        self._respond_result(result)
 
-    def send_request_login(self, record):
+    def send_request_login(self, record: "Record") -> None:
         # TODO: do something with api_key or anonymous?
         # TODO: return an error if we aren't logged in?
         self._api.reauth()
@@ -283,13 +360,13 @@ class SendManager(object):
             logger.info("Login server info: {}".format(server_info))
         self._entity = viewer.get("entity")
         if record.control.req_resp:
-            result = wandb_internal_pb2.Result(uuid=record.uuid)
+            result = proto_util._result_from_record(record)
             if self._entity:
                 result.response.login_response.active_entity = self._entity
-            self._result_q.put(result)
+            self._respond_result(result)
 
-    def send_exit(self, data):
-        exit = data.exit
+    def send_exit(self, record: "Record") -> None:
+        exit = record.exit
         self._exit_code = exit.exit_code
         logger.info("handling exit code: %s", exit.exit_code)
         runtime = exit.runtime
@@ -297,24 +374,20 @@ class SendManager(object):
         self._metadata_summary["runtime"] = runtime
         self._update_summary()
 
-        # Pass the responsibility to respond to handle_request_defer()
-        if data.control.req_resp:
-            self._exit_sync_uuid = data.uuid
-
         # We need to give the request queue a chance to empty between states
         # so use handle_request_defer as a state machine.
         logger.info("send defer")
         self._interface.publish_defer()
 
-    def send_final(self, data):
+    def send_final(self, record: "Record") -> None:
         pass
 
-    def send_request_defer(self, data):
-        defer = data.request.defer
+    def send_request_defer(self, record: "Record") -> None:
+        defer = record.request.defer
         state = defer.state
         logger.info("handle sender defer: {}".format(state))
 
-        def transition_state():
+        def transition_state() -> None:
             state = defer.state + 1
             logger.info("send defer: {}".format(state))
             self._interface.publish_defer(state)
@@ -325,6 +398,9 @@ class SendManager(object):
         elif state == defer.FLUSH_STATS:
             # NOTE: this is handled in handler.py:handle_request_defer()
             transition_state()
+        elif state == defer.FLUSH_PARTIAL_HISTORY:
+            # NOTE: this is handled in handler.py:handle_request_defer()
+            transition_state()
         elif state == defer.FLUSH_TB:
             # NOTE: this is handled in handler.py:handle_request_defer()
             transition_state()
@@ -368,27 +444,14 @@ class SendManager(object):
 
         exit_result = wandb_internal_pb2.RunExitResult()
 
-        # This path is not the prefered method to return exit results
-        # as it could take a long time to flush the file pusher buffers
-        if self._exit_sync_uuid:
-            if self._pusher:
-                # NOTE: This will block until finished
-                self._pusher.print_status()
-                self._pusher.join()
-                self._pusher = None
-            resp = wandb_internal_pb2.Result(
-                exit_result=exit_result, uuid=self._exit_sync_uuid
-            )
-            self._result_q.put(resp)
-
         # mark exit done in case we are polling on exit
         self._exit_result = exit_result
 
-    def send_request_poll_exit(self, record):
+    def send_request_poll_exit(self, record: "Record") -> None:
         if not record.control.req_resp:
             return
 
-        result = wandb_internal_pb2.Result(uuid=record.uuid)
+        result = proto_util._result_from_record(record)
 
         alive = False
         if self._pusher:
@@ -412,9 +475,11 @@ class SendManager(object):
                 self.get_local_info()
             )
             result.response.poll_exit_response.done = True
-        self._result_q.put(result)
+        self._respond_result(result)
 
-    def _maybe_setup_resume(self, run) -> "Optional[wandb_internal_pb2.ErrorInfo]":
+    def _maybe_setup_resume(
+        self, run: "RunRecord"
+    ) -> "Optional[wandb_internal_pb2.ErrorInfo]":
         """This maybe queries the backend for a run and fails if the settings are
         incompatible."""
         if not self._settings.resume:
@@ -467,9 +532,7 @@ class SendManager(object):
             summary = json.loads(resume_status["summaryMetrics"] or "{}")
             new_runtime = summary.get("_wandb", {}).get("runtime", None)
             if new_runtime is not None:
-                if "_wandb" not in self._resume_state:
-                    self._resume_state["_wandb"] = {}
-                self._resume_state["_wandb"].update({"runtime": new_runtime})
+                self._resume_state.wandb_runtime = new_runtime
 
         except (IndexError, ValueError) as e:
             logger.error("unable to load resume tails", exc_info=e)
@@ -481,14 +544,14 @@ class SendManager(object):
 
         # TODO: Do we need to restore config / summary?
         # System metrics runtime is usually greater than history
-        self._resume_state["runtime"] = max(events_rt, history_rt)
-        self._resume_state["step"] = history.get("_step", -1) + 1 if history else 0
-        self._resume_state["history"] = resume_status["historyLineCount"]
-        self._resume_state["events"] = resume_status["eventsLineCount"]
-        self._resume_state["output"] = resume_status["logLineCount"]
-        self._resume_state["config"] = config
-        self._resume_state["summary"] = summary
-        self._resume_state["resumed"] = True
+        self._resume_state.runtime = max(events_rt, history_rt)
+        self._resume_state.step = history.get("_step", -1) + 1 if history else 0
+        self._resume_state.history = resume_status["historyLineCount"]
+        self._resume_state.events = resume_status["eventsLineCount"]
+        self._resume_state.output = resume_status["logLineCount"]
+        self._resume_state.config = config
+        self._resume_state.summary = summary
+        self._resume_state.resumed = True
         logger.info("configured resuming with: %s" % self._resume_state)
         return None
 
@@ -554,26 +617,29 @@ class SendManager(object):
         config_path = os.path.join(self._settings.files_dir, "config.yaml")
         config_util.save_config_file_from_dict(config_path, config_value_dict)
 
-    def _sync_spell(self, env=None):
+    def _sync_spell(self) -> None:
         """Syncs this run with spell"""
+        if not self._run:
+            return
         try:
-            env = env or os.environ
+            env = os.environ
             self._interface.publish_config(
                 key=("_wandb", "spell_url"), val=env.get("SPELL_RUN_URL")
             )
             url = "{}/{}/{}/runs/{}".format(
                 self._api.app_url, self._run.entity, self._run.project, self._run.run_id
             )
-            return requests.put(
+            requests.put(
                 env.get("SPELL_API_URL", "https://api.spell.run") + "/wandb_url",
                 json={"access_token": env.get("WANDB_ACCESS_TOKEN"), "url": url},
                 timeout=2,
             )
         except requests.RequestException:
-            return False
+            pass
+        # TODO: do something if sync spell is not successful?
 
-    def send_run(self, data, file_dir=None) -> None:
-        run = data.run
+    def send_run(self, record: "Record", file_dir: str = None) -> None:
+        run = record.run
         error = None
         is_wandb_init = self._run is None
 
@@ -599,20 +665,20 @@ class SendManager(object):
             error = self._maybe_setup_resume(run)
 
         if error is not None:
-            if data.control.req_resp:
-                resp = wandb_internal_pb2.Result(uuid=data.uuid)
-                resp.run_result.run.CopyFrom(run)
-                resp.run_result.error.CopyFrom(error)
-                self._result_q.put(resp)
+            if record.control.req_resp:
+                result = proto_util._result_from_record(record)
+                result.run_result.run.CopyFrom(run)
+                result.run_result.error.CopyFrom(error)
+                self._respond_result(result)
             else:
                 logger.error("Got error in async mode: %s", error.message)
             return
 
         # Save the resumed config
-        if self._resume_state["config"] is not None:
+        if self._resume_state.config is not None:
             # TODO: should we merge this with resumed config?
             config_override = self._consolidated_config
-            config_dict = self._resume_state["config"]
+            config_dict = self._resume_state.config
             config_dict = config_util.dict_strip_value_dict(config_dict)
             config_dict.update(config_override)
             self._consolidated_config.update(config_dict)
@@ -627,14 +693,15 @@ class SendManager(object):
             self._config_save(config_value_dict)
 
         self._init_run(run, config_value_dict)
+        assert self._run  # self._run is configured in _init_run()
 
-        if data.control.req_resp:
-            resp = wandb_internal_pb2.Result(uuid=data.uuid)
+        if record.control.req_resp:
+            result = proto_util._result_from_record(record)
             # TODO: we could do self._interface.publish_defer(resp) to notify
             # the handler not to actually perform server updates for this uuid
             # because the user process will send a summary update when we resume
-            resp.run_result.run.CopyFrom(self._run)
-            self._result_q.put(resp)
+            result.run_result.run.CopyFrom(self._run)
+            self._respond_result(result)
 
         # Only spin up our threads on the first run message
         if is_wandb_init:
@@ -642,10 +709,11 @@ class SendManager(object):
         else:
             logger.info("updated run: %s", self._run.run_id)
 
-    def _init_run(self, run, config_dict):
+    def _init_run(
+        self, run: "RunRecord", config_dict: Optional[DictWithValues]
+    ) -> None:
         # We subtract the previous runs runtime when resuming
-        start_time = run.start_time.ToSeconds() - self._resume_state["runtime"]
-        repo = GitRepo(remote=self._settings.git_remote)
+        start_time = run.start_time.ToSeconds() - self._resume_state.runtime
         # TODO: we don't check inserted currently, ultimately we should make
         # the upsert know the resume state and fail transactionally
         server_run, inserted = self._api.upsert_run(
@@ -661,23 +729,20 @@ class SendManager(object):
             sweep_name=run.sweep_id or None,
             host=run.host or None,
             program_path=self._settings.program or None,
-            repo=repo.remote_url,
-            commit=repo.last_commit,
+            repo=run.git.remote_url or None,
+            commit=run.git.last_commit or None,
         )
         self._run = run
-        if self._resume_state.get("resumed"):
+        if self._resume_state.resumed:
             self._run.resumed = True
-            if (
-                "_wandb" in self._resume_state
-                and "runtime" in self._resume_state["_wandb"]
-            ):
-                self._run.runtime = self._resume_state["_wandb"]["runtime"]
-        self._run.starting_step = self._resume_state["step"]
+            if self._resume_state.wandb_runtime is not None:
+                self._run.runtime = self._resume_state.wandb_runtime
+        self._run.starting_step = self._resume_state.step
         self._run.start_time.FromSeconds(int(start_time))
         self._run.config.CopyFrom(self._interface._make_config(config_dict))
-        if self._resume_state["summary"] is not None:
+        if self._resume_state.summary is not None:
             self._run.summary.CopyFrom(
-                self._interface._make_summary_from_dict(self._resume_state["summary"])
+                self._interface._make_summary_from_dict(self._resume_state.summary)
             )
         storage_id = server_run.get("id")
         if storage_id:
@@ -711,7 +776,8 @@ class SendManager(object):
         if os.getenv("SPELL_RUN_URL"):
             self._sync_spell()
 
-    def _start_run_threads(self, file_dir=None):
+    def _start_run_threads(self, file_dir: str = None) -> None:
+        assert self._run  # self._run is configured by caller
         self._fs = file_stream.FileStreamApi(
             self._api,
             self._run.run_id,
@@ -722,22 +788,22 @@ class SendManager(object):
         self._fs.set_file_policy("wandb-summary.json", file_stream.SummaryFilePolicy())
         self._fs.set_file_policy(
             "wandb-history.jsonl",
-            file_stream.JsonlFilePolicy(start_chunk_id=self._resume_state["history"]),
+            file_stream.JsonlFilePolicy(start_chunk_id=self._resume_state.history),
         )
         self._fs.set_file_policy(
             "wandb-events.jsonl",
-            file_stream.JsonlFilePolicy(start_chunk_id=self._resume_state["events"]),
+            file_stream.JsonlFilePolicy(start_chunk_id=self._resume_state.events),
         )
         self._fs.set_file_policy(
             "output.log",
-            file_stream.CRDedupeFilePolicy(start_chunk_id=self._resume_state["output"]),
-        )
-        util.sentry_set_scope(
-            "internal",
-            entity=self._run.entity,
-            project=self._run.project,
-            email=self._settings.email,
+            file_stream.CRDedupeFilePolicy(start_chunk_id=self._resume_state.output),
         )
+
+        # hack to merge run_settings and self._settings object together
+        # so that fields like entity or project are available to be attached to Sentry events.
+        run_settings = message_to_dict(self._run)
+        self._settings = SettingsStatic({**dict(self._settings), **run_settings})
+        util.sentry_set_scope(settings_dict=self._settings,)
         self._fs.start()
         self._pusher = FilePusher(self._api, self._fs, silent=self._settings.silent)
         self._dir_watcher = DirWatcher(
@@ -749,21 +815,21 @@ class SendManager(object):
             self._run.start_time.ToSeconds(),
         )
 
-    def _save_history(self, history_dict):
+    def _save_history(self, history_dict: Dict[str, Any]) -> None:
         if self._fs:
             self._fs.push(filenames.HISTORY_FNAME, json.dumps(history_dict))
 
-    def send_history(self, data):
-        history = data.history
+    def send_history(self, record: "Record") -> None:
+        history = record.history
         history_dict = proto_util.dict_from_proto_list(history.item)
         self._save_history(history_dict)
 
-    def send_summary(self, data):
-        summary_dict = proto_util.dict_from_proto_list(data.summary.update)
+    def send_summary(self, record: "Record") -> None:
+        summary_dict = proto_util.dict_from_proto_list(record.summary.update)
         self._cached_summary = summary_dict
         self._update_summary()
 
-    def _update_summary(self):
+    def _update_summary(self) -> None:
         summary_dict = self._cached_summary.copy()
         summary_dict.pop("_wandb", None)
         if self._metadata_summary:
@@ -777,17 +843,19 @@ class SendManager(object):
             f.write(json_summary)
         self._save_file(filenames.SUMMARY_FNAME)
 
-    def send_stats(self, data):
-        stats = data.stats
+    def send_stats(self, record: "Record") -> None:
+        stats = record.stats
         if stats.stats_type != wandb_internal_pb2.StatsRecord.StatsType.SYSTEM:
             return
         if not self._fs:
             return
+        if not self._run:
+            return
         now = stats.timestamp.seconds
         d = dict()
         for item in stats.item:
             d[item.key] = json.loads(item.value_json)
-        row = dict(system=d)
+        row: Dict[str, Any] = dict(system=d)
         self._flatten(row)
         row["_wandb"] = True
         row["_timestamp"] = now
@@ -795,10 +863,10 @@ class SendManager(object):
         self._fs.push(filenames.EVENTS_FNAME, json.dumps(row))
         # TODO(jhr): check fs.push results?
 
-    def send_output(self, data):
+    def send_output(self, record: "Record") -> None:
         if not self._fs:
             return
-        out = data.output
+        out = record.output
         prepend = ""
         stream = "stdout"
         if out.output_type == wandb_internal_pb2.OutputRecord.OutputType.STDERR:
@@ -819,20 +887,20 @@ class SendManager(object):
             cur_time = time.time()
             timestamp = datetime.utcfromtimestamp(cur_time).isoformat() + " "
             prev_str = self._partial_output.get(stream, "")
-            line = u"{}{}{}{}".format(prepend, timestamp, prev_str, line)
+            line = "{}{}{}{}".format(prepend, timestamp, prev_str, line)
             self._fs.push(filenames.OUTPUT_FNAME, line)
             self._partial_output[stream] = ""
 
-    def _update_config(self):
+    def _update_config(self) -> None:
         self._config_needs_debounce = True
 
-    def send_config(self, data):
-        cfg = data.config
+    def send_config(self, record: "Record") -> None:
+        cfg = record.config
         config_util.update_from_proto(self._consolidated_config, cfg)
         self._update_config()
 
-    def send_metric(self, data: wandb_internal_pb2.Record) -> None:
-        metric = data.metric
+    def send_metric(self, record: "Record") -> None:
+        metric = record.metric
         if metric.glob_name:
             logger.warning("Seen metric with glob (shouldnt happen)")
             return
@@ -870,50 +938,73 @@ class SendManager(object):
             self._config_metric_index_dict[metric.name] = next_idx
         self._update_config()
 
-    def send_telemetry(self, data):
-        telem = data.telemetry
+    def send_telemetry(self, record: "Record") -> None:
+        telem = record.telemetry
         self._telemetry_obj.MergeFrom(telem)
         self._update_config()
 
-    def _save_file(self, fname, policy="end"):
+    def _save_file(self, fname: str, policy: str = "end") -> None:
         logger.info("saving file %s with policy %s", fname, policy)
         if self._dir_watcher:
             self._dir_watcher.update_policy(fname, policy)
 
-    def send_files(self, data):
-        files = data.files
+    def send_files(self, record: "Record") -> None:
+        files = record.files
         for k in files.files:
             # TODO(jhr): fix paths with directories
             self._save_file(k.path, interface.file_enum_to_policy(k.policy))
 
-    def send_header(self, data):
+    def send_header(self, record: "Record") -> None:
         pass
 
-    def send_footer(self, data):
+    def send_footer(self, record: "Record") -> None:
         pass
 
-    def send_tbrecord(self, data):
+    def send_tbrecord(self, record: "Record") -> None:
         # tbrecord watching threads are handled by handler.py
         pass
 
-    def send_request_log_artifact(self, record):
+    def send_request_log_artifact(self, record: "Record") -> None:
         assert record.control.req_resp
-        result = wandb_internal_pb2.Result(uuid=record.uuid)
+        result = proto_util._result_from_record(record)
         artifact = record.request.log_artifact.artifact
 
         try:
             res = self._send_artifact(artifact)
-            result.response.log_artifact_response.artifact_id = res.get("id")
+            assert res, "Unable to send artifact"
+            result.response.log_artifact_response.artifact_id = res["id"]
             logger.info("logged artifact {} - {}".format(artifact.name, res))
         except Exception as e:
             result.response.log_artifact_response.error_message = 'error logging artifact "{}/{}": {}'.format(
                 artifact.type, artifact.name, e
             )
 
-        self._result_q.put(result)
+        self._respond_result(result)
 
-    def send_artifact(self, data):
-        artifact = data.artifact
+    def send_request_artifact_send(self, record: "Record") -> None:
+        # TODO: combine and eventually remove send_request_log_artifact()
+
+        # for now we are using req/resp uuid for transaction id
+        # in the future this should be part of the message to handle idempotency
+        xid = record.uuid
+
+        done_msg = wandb_internal_pb2.ArtifactDoneRequest(xid=xid)
+        artifact = record.request.artifact_send.artifact
+        try:
+            res = self._send_artifact(artifact)
+            assert res, "Unable to send artifact"
+            done_msg.artifact_id = res["id"]
+            logger.info("logged artifact {} - {}".format(artifact.name, res))
+        except Exception as e:
+            done_msg.error_message = 'error logging artifact "{}/{}": {}'.format(
+                artifact.type, artifact.name, e
+            )
+
+        logger.info("send artifact done")
+        self._interface._publish_artifact_done(done_msg)
+
+    def send_artifact(self, record: "Record") -> None:
+        artifact = record.artifact
         try:
             res = self._send_artifact(artifact)
             logger.info("sent artifact {} - {}".format(artifact.name, res))
@@ -924,7 +1015,8 @@ class SendManager(object):
                 )
             )
 
-    def _send_artifact(self, artifact):
+    def _send_artifact(self, artifact: "ArtifactRecord") -> Optional[Dict]:
+        assert self._pusher
         saver = artifacts.ArtifactSaver(
             api=self._api,
             digest=artifact.digest,
@@ -942,7 +1034,7 @@ class SendManager(object):
                     "This W&B server doesn't support distributed artifacts, "
                     "have your administrator install wandb/local >= 0.9.37"
                 )
-                return
+                return None
 
         metadata = json.loads(artifact.metadata) if artifact.metadata else None
         return saver.save(
@@ -959,8 +1051,8 @@ class SendManager(object):
             incremental=artifact.incremental_beta1,
         )
 
-    def send_alert(self, data):
-        alert = data.alert
+    def send_alert(self, record: "Record") -> None:
+        alert = record.alert
         max_cli_version = self._max_cli_version()
         if max_cli_version is None or parse_version(max_cli_version) < parse_version(
             "0.10.9"
@@ -982,7 +1074,7 @@ class SendManager(object):
                     'send_alert: failed for alert "{}": {}'.format(alert.title, e)
                 )
 
-    def finish(self):
+    def finish(self) -> None:
         logger.info("shutting down sender")
         # if self._tb_watcher:
         #     self._tb_watcher.finish()
@@ -997,31 +1089,31 @@ class SendManager(object):
             self._fs.finish(self._exit_code)
             self._fs = None
 
-    def _max_cli_version(self):
+    def _max_cli_version(self) -> Optional[str]:
         server_info = self.get_server_info()
         max_cli_version = server_info.get("cliVersionInfo", {}).get(
             "max_cli_version", None
         )
+        if not isinstance(max_cli_version, str):
+            return None
         return max_cli_version
 
-    def get_viewer_server_info(self):
-        if not self._cached_server_info or not self._cached_viewer:
-            (
-                self._cached_viewer,
-                self._cached_server_info,
-            ) = self._api.viewer_server_info()
+    def get_viewer_server_info(self) -> None:
+        if self._cached_server_info and self._cached_viewer:
+            return
+        self._cached_viewer, self._cached_server_info = self._api.viewer_server_info()
 
-    def get_viewer_info(self):
+    def get_viewer_info(self) -> Dict[str, Any]:
         if not self._cached_viewer:
             self.get_viewer_server_info()
         return self._cached_viewer
 
-    def get_server_info(self):
+    def get_server_info(self) -> Dict[str, Any]:
         if not self._cached_server_info:
             self.get_viewer_server_info()
         return self._cached_server_info
 
-    def get_local_info(self):
+    def get_local_info(self) -> "LocalInfo":
         """
         This is a helper function that queries the server to get the the local version information.
         First, we perform an introspection, if it returns empty we deduce that the docker image is
@@ -1048,7 +1140,7 @@ class SendManager(object):
             )
         return local_info
 
-    def __next__(self):
+    def __next__(self) -> "Record":
         return self._record_q.get(block=True)
 
     next = __next__
