@@ -4,15 +4,12 @@
 static settings.
 """
 
-from typing import TYPE_CHECKING
+from typing import Any, Dict, Iterable, Optional, Tuple, Union
 
-if TYPE_CHECKING:
-    from typing import Any, Dict, Iterable, Optional, Union
-
-    SettingsDict = Dict[str, Union[str, float]]
+SettingsDict = Dict[str, Union[str, float, Tuple, None]]
 
 
-class SettingsStatic(object):
+class SettingsStatic:
     # TODO(jhr): figure out how to share type defs with sdk/wandb_settings.py
     _offline: "Optional[bool]"
     _disable_stats: "Optional[bool]"
@@ -21,6 +18,14 @@ class SettingsStatic(object):
     files_dir: str
     log_internal: str
     _internal_check_process: bool
+    is_local: "Optional[bool]"
+    _colab: "Optional[bool]"
+    _jupyter: "Optional[bool]"
+    _require_service: "Optional[str]"
+    resume: "Optional[str]"
+    program: "Optional[str]"
+    silent: "Optional[bool]"
+    email: "Optional[str]"
 
     # TODO(jhr): clean this up, it is only in SettingsStatic and not in Settings
     _log_level: int
