@@ -8,7 +8,12 @@ import logging
 
 import six
 import wandb
-from wandb.util import json_friendly_val
+from wandb.util import (
+    _is_artifact,
+    _is_artifact_string,
+    check_dict_contains_nested_artifact,
+    json_friendly_val,
+)
 
 from . import wandb_helper
 from .lib import config_util
@@ -96,12 +101,16 @@ class Config(object):
         object.__setattr__(self, "_users_cnt", 0)
         object.__setattr__(self, "_callback", None)
         object.__setattr__(self, "_settings", None)
+        object.__setattr__(self, "_artifact_callback", None)
 
         self._load_defaults()
 
     def _set_callback(self, cb):
         object.__setattr__(self, "_callback", cb)
 
+    def _set_artifact_callback(self, cb):
+        object.__setattr__(self, "_artifact_callback", cb)
+
     def _set_settings(self, settings):
         object.__setattr__(self, "_settings", settings)
 
@@ -136,8 +145,9 @@ class Config(object):
     def __setitem__(self, key, val):
         if self._check_locked(key):
             return
-        with wandb.wandb_lib.telemetry.context() as tel:
+        with wandb.sdk.lib.telemetry.context() as tel:
             tel.feature.set_config_item = True
+        self._raise_value_error_on_nested_artifact(val, nested=True)
         key, val = self._sanitize(key, val)
         self._items[key] = val
         logger.info("config set %s = %s - %s", key, val, self._callback)
@@ -211,9 +221,10 @@ class Config(object):
             self.update(conf_dict)
 
     def _sanitize_dict(
-        self, config_dict, allow_val_change=None, ignore_keys: set = None
+        self, config_dict, allow_val_change=None, ignore_keys: set = None,
     ):
         sanitized = {}
+        self._raise_value_error_on_nested_artifact(config_dict)
         for k, v in six.iteritems(config_dict):
             if ignore_keys and k in ignore_keys:
                 continue
@@ -227,7 +238,14 @@ class Config(object):
             allow_val_change = True
         # We always normalize keys by stripping '-'
         key = key.strip("-")
-        val = json_friendly_val(val)
+        if _is_artifact_string(val) or _is_artifact(val):
+            val = self._artifact_callback(key, val)
+        # if the user inserts an artifact into the config
+        if not (
+            isinstance(val, wandb.Artifact)
+            or isinstance(val, wandb.apis.public.Artifact)
+        ):
+            val = json_friendly_val(val)
         if not allow_val_change:
             if key in self._items and val != self._items[key]:
                 raise config_util.ConfigError(
@@ -240,6 +258,15 @@ class Config(object):
                 )
         return key, val
 
+    def _raise_value_error_on_nested_artifact(self, v, nested=False):
+        # we can't swap nested artifacts because their root key can be locked by other values
+        # best if we don't allow nested artifacts until we can lock nested keys in the config
+        if isinstance(v, dict) and check_dict_contains_nested_artifact(v, nested):
+            raise ValueError(
+                "Instances of wandb.Artifact and wandb.apis.public.Artifact"
+                " can only be top level keys in wandb.config"
+            )
+
 
 class ConfigStatic(object):
     def __init__(self, config):
